/*
 * main.cxx
 *
 * PWLib application source file for opengk
 *
 * Main program entry point.
 *
 * Copyright 2001 Equivalence
 *
 * $Log: main.cxx,v $
 * Revision 1.41  2006/04/27 05:46:31  rjongbloed
 * Changed so RIP is sent for slow ARQ response when OSP is used only, not
 *   all the time as it was which caused error in simple gatekeepers.
 *
 * Revision 1.40  2006/04/25 09:10:04  rjongbloed
 * Ported gatekeeper server to (optionally) use OPAL code base
 *
 * Revision 1.39  2005/03/11 00:20:27  csoutheren
 * Fixed problem when compiling without H.501 support
 *
 * Revision 1.38  2004/12/18 03:41:33  rjongbloed
 * Fixed being able to compile without OSP library
 *
 * Revision 1.37  2004/12/16 00:47:55  csoutheren
 * Added incoming call functions
 *
 * Revision 1.36  2004/12/15 06:29:53  csoutheren
 * Added OSP support
 *
 * Revision 1.35  2004/12/14 11:15:56  rjongbloed
 * Added extra parameter to TranslateAliasAddress for OSP support
 *
 * Revision 1.34  2004/08/15 07:19:26  rjongbloed
 * Fixed compatibility with gkserver changes
 *
 * Revision 1.33  2004/05/03 13:25:46  rjongbloed
 * Converted everything to be codec plug in freindly
 * Removed GSM and G.729 as now plug ins are "the way"!
 *
 * Revision 1.32  2003/04/02 07:06:17  robertj
 * Added support for (optional) duplicate aliases.
 *
 * Revision 1.31  2003/04/02 06:13:36  robertj
 * Added H.323 Annex G Peer Element support to gatekeeper server.
 *
 * Revision 1.30  2003/04/01 07:30:32  robertj
 * Added function to translate alias into alias list and transport to make
 *   use of full information returned by peer element.
 *
 * Revision 1.29  2003/03/27 09:24:47  craigs
 * Changed to support return of multiple aliases from PE
 *
 * Revision 1.28  2003/03/25 05:14:51  craigs
 * Changed to return RIP on ARQ due to slow lookups
 *
 * Revision 1.27  2003/03/25 02:26:18  robertj
 * Added ability to adjust call heartbeat time
 *
 * Revision 1.26  2003/03/25 01:43:18  craigs
 * Changes for new OpenH323 H.501
 *
 * Revision 1.25  2003/03/18 13:59:57  craigs
 * Changed to keep trying to find clearing house after startup
 *
 * Revision 1.24  2003/03/18 00:01:11  craigs
 * Added support for H.501 clearing houses
 *
 * Revision 1.23  2003/02/02 01:47:54  robertj
 * Added ability to set static route maps for aliases to gateway hosts.
 *
 * Revision 1.22  2002/11/28 05:59:43  robertj
 * Fixed bug so can set total bandwidth while calls are in progress.
 *
 * Revision 1.21  2002/11/01 06:42:48  robertj
 * Changed status display to be UTF-8
 *
 * Revision 1.20  2002/10/21 03:12:26  robertj
 * Added unregister an endpoint or clear a call buttons to status page.
 *
 * Revision 1.19  2002/07/17 09:19:24  robertj
 * Allowed for not having gif file for product name in default header.
 * Changed to use default pwlib page graphic after above changes.
 *
 * Revision 1.18  2002/07/17 04:01:30  robertj
 * Fixed removal of password entries.
 *
 * Revision 1.17  2002/07/12 02:27:51  robertj
 * Added more call related info to status page.
 *
 * Revision 1.16  2002/06/19 05:04:58  robertj
 * Changes gk code to allow for H.235 security on an endpoint by endpoint basis
 *   caused slight semantic change to H.235 flag.
 *
 * Revision 1.15  2002/06/12 03:55:54  robertj
 * Added function to add/remove multiple listeners in one go comparing against
 *   what is already running so does not interrupt unchanged listeners.
 *
 * Revision 1.14  2002/05/31 08:33:16  robertj
 * Changed test oids to Equivalence private numbers
 *
 * Revision 1.13  2002/05/29 07:10:56  robertj
 * Added test of tokens in ACF.
 * Added http links to get at trace log file (if available).
 *
 * Revision 1.12  2002/05/07 06:51:23  robertj
 * Added application name to registered endpoint info.
 *
 * Revision 1.11  2002/05/07 06:16:58  robertj
 * Changed ARQ processing so rejects multiple ARQ with same call id, not this
 *   does NOT include the transport level retries (based on sequence number).
 *
 * Revision 1.10  2002/05/06 00:49:06  robertj
 * Sizeable rewrite of gatekeeper server code to make more bulletproof against
 *   multithreaded operation. Especially when using slow response/RIP feature.
 * Also changed the call indexing to use call id and direction as key.
 *
 * Revision 1.9  2002/03/01 05:37:35  robertj
 * Changed status call display to show both endpoints in call.
 *
 * Revision 1.8  2002/02/25 10:56:05  robertj
 * Added dynamically allocated port for H.323 channel listener.
 *
 * Revision 1.7  2002/02/17 23:51:18  robertj
 * Changed default listener port for future gk routing to 1721 instead of default.
 *
 * Revision 1.6  2002/02/04 05:20:03  robertj
 * Added call status display on HTTP.
 * Added test for slow ARQ
 *
 * Revision 1.5  2002/01/17 07:03:29  robertj
 * Added support for RFC2833 embedded DTMF in the RTP stream
 *
 * Revision 1.4  2001/12/22 04:33:54  robertj
 * Fixed ability to change listener port, thanks george.2.hare@bt.com
 *
 * Revision 1.3  2001/09/18 09:55:08  robertj
 * Changed default to have no username and password.
 *
 * Revision 1.2  2001/08/10 11:03:49  robertj
 * Major changes to H.235 support in RAS to support server.
 *
 * Revision 1.1  2001/08/03 04:48:57  robertj
 * First cut of a gatekeeper
 *
 */

#include "main.h"
#include "custom.h"

#include <ptclib/random.h>

#ifdef H323_TRANSNEXUS_OSP
#include <opalosp.h>
#endif

//#define TEST_TOKEN

PCREATE_PROCESS(OpenGk);

const WORD DefaultHTTPPort = 1719;

static const char LogLevelKey[] = "Log Level";
static const char AdminUserNameKey[] = "Username";
static const char AdminPasswordKey[] = "Password";
static const char HttpPortKey[] = "HTTP Port";
static const char LocalUserNameKey[] = "Local User Name";
static const char TypeOfServiceKey[] = "Type Of Service";
static const char GatekeeperIdentifierKey[] = "Gatekeeper Identifier";
static const char AvailableBandwidthKey[] = "Total Bandwidth";
static const char DefaultBandwidthKey[] = "Default Bandwidth Allocation";
static const char MaximumBandwidthKey[] = "Maximum Bandwidth Allocation";
static const char DefaultTimeToLiveKey[] = "Default Time To Live";
static const char CallHeartbeatTimeKey[]    = "Call Heartbeat Time";
static const char OverwriteOnSameSignalAddressKey[] = "Overwrite EP On Same Signal Address";
static const char CanHaveDuplicateAliasKey[] = "Can Have Duplicate Alias";
static const char CanOnlyCallRegisteredEPKey[] = "Can Only Call Registered EP";
static const char CanOnlyAnswerRegisteredEPKey[] = "Can Only Answer Registered EP";
static const char AnswerCallPreGrantedARQKey[] = "Answer Call Pregranted ARQ";
static const char MakeCallPreGrantedARQKey[] = "Make Call Pregranted ARQ";
static const char IsGatekeeperRoutedKey[] = "Gatekeeper Routed";
static const char AliasCanBeHostNameKey[] = "Alias Can Be Host Name";
static const char RequireH235Key[] = "Require H.235";
static const char UsernameKey[] = "Username";
static const char PasswordKey[] = "Password";
static const char RouteAliasKey[] = "Alias";
static const char RouteHostKey[] = "Host";
static const char ClearingHouseKey[] = "H501 Clearing House";
static const char H501InterfaceKey[] = "H501 Interface";

#ifdef H323_TRANSNEXUS_OSP
static const char OSPRoutingURLKey[] = "OSP Routing URL";
static const char OSPPrivateKeyFileKey[] = "OSP Private Key";
static const char OSPPublicKeyFileKey[] = "OSP Public Key";
static const char OSPServerKeyFileKey[] = "OSP Server Key";
#endif

#define GATEWAY_LISTENERS_SECTION     "Gateway Listeners"
#define GATEKEEPER_LISTENERS_SECTION  "Gatekeeper Listeners"
#define LISTENER_INTERFACE_KEY        "Interface"

#define USERS_SECTION "Authentication"
#define USERS_KEY     "Credentials"

#define ROUTES_SECTION "Route Maps"
#define ROUTES_KEY     "Mapping"


#define new PNEW


///////////////////////////////////////////////////////////////

OpenGk::OpenGk()
  : PHTTPServiceProcess(ProductInfo)
{
  endpoint = NULL;
}


bool OpenGk::OnStart()
{
  // Create the endpoint, or reinitialised if already exists
  endpoint = new MyEndPoint(
#ifdef USE_OPAL
    manager
#endif
  );

  // change to the default directory to the one containing the executable
  PDirectory exeDir = GetFile().GetDirectory();

#if defined(_WIN32) && defined(_DEBUG)
  // Special check to aid in using DevStudio for debugging.
  if (exeDir.Find("\\Debug\\") != P_MAX_INDEX)
    exeDir = exeDir.GetParent();
#endif
  exeDir.Change();

  httpNameSpace.AddResource(new PHTTPDirectory("data", "data"));
  httpNameSpace.AddResource(new PServiceHTTPDirectory("html", "html"));

  return PHTTPServiceProcess::OnStart();
}


void OpenGk::OnStop()
{
  delete endpoint;
  endpoint = NULL;

  PHTTPServiceProcess::OnStop();
}



void OpenGk::OnControl()
{
  // This function get called when the Control menu item is selected in the
  // tray icon mode of the service.
  PStringStream url;
  url << "http://";

  PString host = PIPSocket::GetHostName();
  PIPSocket::Address addr;
  if (PIPSocket::GetHostAddress(host, addr))
    url << host;
  else
    url << "localhost";

  url << ':' << DefaultHTTPPort;

  PURL::OpenBrowser(url);
}


void OpenGk::OnConfigChanged()
{
}


bool OpenGk::Initialise(const char * initMsg)
{
  PConfig cfg("Parameters");

  // Sert log level as early as possible
  SetLogLevel((PSystemLog::Level)cfg.GetInteger(LogLevelKey, GetLogLevel()));
#if PTRACING
  if (GetLogLevel() >= PSystemLog::Warning)
    PTrace::SetLevel(GetLogLevel()-PSystemLog::Warning);
  else
    PTrace::SetLevel(0);
  PTrace::ClearOptions(PTrace::Timestamp);
  PTrace::SetOptions(PTrace::DateAndTime);
#endif

  // Get the HTTP basic authentication info
  PString adminUserName = cfg.GetString(AdminUserNameKey);
  PString adminPassword = PHTTPPasswordField::Decrypt(cfg.GetString(AdminPasswordKey));

  PHTTPSimpleAuth authority(GetName(), adminUserName, adminPassword);


  // Create the parameters URL page, and start adding fields to it
  PConfigPage * rsrc = new PConfigPage(*this, "Parameters", "Parameters", authority);

  // HTTP authentication username/password
  rsrc->Add(new PHTTPStringField(AdminUserNameKey, 25, adminUserName));
  rsrc->Add(new PHTTPPasswordField(AdminPasswordKey, 25, adminPassword));

  // Log level for messages
  rsrc->Add(new PHTTPIntegerField(LogLevelKey,
                                  PSystemLog::Fatal, PSystemLog::NumLogLevels-1,
                                  GetLogLevel(),
                                  "1=Fatal only, 2=Errors, 3=Warnings, 4=Info, 5=Debug"));

  // HTTP Port number to use.
  WORD httpPort = (WORD)cfg.GetInteger(HttpPortKey, DefaultHTTPPort);
  rsrc->Add(new PHTTPIntegerField(HttpPortKey, 1, 32767, httpPort));

  // Initialise the rest of the endpoint parameters
  endpoint->Initialise(cfg, rsrc);

  // Finished the resource to add, generate HTML for it and add to name space
  PServiceHTML html("System Parameters");
  rsrc->BuildHTML(html);
  httpNameSpace.AddResource(rsrc, PHTTPSpace::Overwrite);


  // Create the status page
  httpNameSpace.AddResource(new MainStatusPage(*this, authority), PHTTPSpace::Overwrite);


  // Add log file links
  if (!systemLogFileName && systemLogFileName != "-") {
    httpNameSpace.AddResource(new PHTTPFile("logfile.txt", systemLogFileName, authority));
    httpNameSpace.AddResource(new PHTTPTailFile("tail_logfile", systemLogFileName, authority));
  }


  //  create the home page
  static const char welcomeHtml[] = "welcome.html";
  if (PFile::Exists(welcomeHtml))
    httpNameSpace.AddResource(new PServiceHTTPFile(welcomeHtml, TRUE), PHTTPSpace::Overwrite);
  else {
    PHTML html;
    html << PHTML::Title("Welcome to OpenH323 Gatekeeper")
         << PHTML::Body()
         << GetPageGraphic()
         << PHTML::Paragraph() << "<center>"

         << PHTML::HotLink("Parameters") << "Parameters" << PHTML::HotLink()
         << PHTML::Paragraph()
         << PHTML::HotLink("Status") << "Status" << PHTML::HotLink()
         << PHTML::Paragraph();

    if (!systemLogFileName && systemLogFileName != "-")
      html << PHTML::HotLink("logfile.txt") << "Full Log File" << PHTML::HotLink()
           << PHTML::BreakLine()
           << PHTML::HotLink("tail_logfile") << "Tail Log File" << PHTML::HotLink()
           << PHTML::Paragraph();
 
    html << PHTML::HRule()
         << GetCopyrightText()
         << PHTML::Body();
    httpNameSpace.AddResource(new PServiceHTTPString("welcome.html", html), PHTTPSpace::Overwrite);
  }

  // set up the HTTP port for listening & start the first HTTP thread
  if (ListenForHTTP(httpPort))
    PSYSTEMLOG(Info, "Opened master socket for HTTP: " << httpListeningSocket->GetPort());
  else {
    PSYSTEMLOG(Fatal, "Cannot run without HTTP port: " << httpListeningSocket->GetErrorText());
    return FALSE;
  }

  PSYSTEMLOG(Info, "Service " << GetName() << ' ' << initMsg);
  return TRUE;
}


void OpenGk::Main()
{
  Suspend();
}


///////////////////////////////////////////////////////////////

MainStatusPage::MainStatusPage(OpenGk & _app, PHTTPAuthority & auth)
  : PServiceHTTPString("Status", "", "text/html; charset=UTF-8", auth),
    app(_app)
{
  PHTML html;

  html << PHTML::Title("OpenH323 Gatekeeper Status")
       << "<meta http-equiv=\"Refresh\" content=\"30\">\n"
       << PHTML::Body()
       << app.GetPageGraphic()
       << PHTML::Paragraph() << "<center>"

       << PHTML::Form("POST")

       << PHTML::TableStart("border=1")
       << PHTML::TableRow()
       << PHTML::TableHeader()
       << "&nbsp;End&nbsp;Point&nbsp;Identifier&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Call&nbsp;Signal&nbsp;Addresses&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Aliases&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Application&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Active&nbsp;Calls&nbsp;"
       << "<!--#macrostart EndPointStatus-->"
         << PHTML::TableRow()
         << PHTML::TableData()
         << "<!--#status EndPointIdentifier-->"
         << PHTML::TableData()
         << "<!--#status CallSignalAddresses-->"
         << PHTML::TableData("NOWRAP")
         << "<!--#status EndPointAliases-->"
         << PHTML::TableData("NOWRAP")
         << "<!--#status Application-->"
         << PHTML::TableData("align=center")
         << "<!--#status ActiveCalls-->"
         << PHTML::TableData()
         << PHTML::SubmitButton("Unregister", "!--#status EndPointIdentifier--")
       << "<!--#macroend EndPointStatus-->"
       << PHTML::TableEnd()

       << PHTML::Paragraph()

       << PHTML::TableStart("border=1")
       << PHTML::TableRow()
       << PHTML::TableHeader()
       << "&nbsp;Call&nbsp;Identifier&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;End&nbsp;Point&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Source/Destination" << PHTML::BreakLine() << "Signalling&nbsp;Addresse&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Last&nbsp;IRR&nbsp;"
       << PHTML::TableHeader()
       << "&nbsp;Connected&nbsp;"
       << "<!--#macrostart CallStatus-->"
         << PHTML::TableRow()
         << PHTML::TableData("NOWRAP")
         << "<!--#status CallIdentifier-->"
         << PHTML::TableData("NOWRAP")
         << "<!--#status EndPointIdentifier-->"
         << PHTML::TableData("NOWRAP")
         << "<!--#status SourceAddress-->"
         << PHTML::BreakLine()
         << "<!--#status DestinationAddress-->"
         << PHTML::TableData()
         << "<!--#status LastIRR-->"
         << PHTML::TableData()
         << "<!--#status ConnectedTime-->"
         << PHTML::TableData()
         << PHTML::SubmitButton("Clear", "!--#status CallIdentifier--")
       << "<!--#macroend CallStatus-->"
       << PHTML::TableEnd()

       << PHTML::Form()
       << PHTML::HRule()

       << app.GetCopyrightText()
       << PHTML::Body();

  string = html;
}


bool MainStatusPage::Post(PHTTPRequest & request,
                          const PStringToString & data,
                          PHTML & msg)
{
  PTRACE(2, "Main\tClear call POST received " << data);

  msg << PHTML::Title() << "Accepted Control Command" << PHTML::Body()
      << PHTML::Heading(1) << "Accepted Control Command" << PHTML::Heading(1);

  if (!app.GetEndPoint().OnPostControl(data, msg))
    msg << PHTML::Heading(2) << "No calls or endpoints!" << PHTML::Heading(2);

  msg << PHTML::Paragraph()
      << PHTML::HotLink(request.url.AsString()) << "Reload page" << PHTML::HotLink()
      << "&nbsp;&nbsp;&nbsp;&nbsp;"
      << PHTML::HotLink("/") << "Home page" << PHTML::HotLink();

  PServiceHTML::ProcessMacros(request, msg, "html/status.html",
                              PServiceHTML::LoadFromFile|PServiceHTML::NoSignatureForFile);
  return TRUE;
}


bool MyEndPoint::OnPostControl(const PStringToString & data, PHTML & msg)
{
  bool gotOne = FALSE;

  for (PINDEX i = 0; i < data.GetSize(); i++) {
    PString key = data.GetKeyAt(i);
    PString value = data.GetDataAt(i);
    if (value == "Unregister") {
      PSafePtr<H323RegisteredEndPoint> ep = gatekeeperServer.FindEndPointByIdentifier(key);
      if (ep != NULL) {
        msg << PHTML::Heading(2) << "Unregistered endpoint " << *ep << PHTML::Heading(2);
        ep->Unregister();
        gotOne = TRUE;
      }
    }
    else if (value == "Clear") {
      PSafePtr<H323GatekeeperCall> call = gatekeeperServer.FindCall(key);
      if (call != NULL) {
        msg << PHTML::Heading(2) << "Cleared call " << *call << PHTML::Heading(2);
        call->Disengage();
        gotOne = TRUE;
      }
    }
  }

  return gotOne;
}


static void SpliceMacro(PString & text, const PString & token, const PString & value)
{
  PRegularExpression RegEx("<?!--#status[ \t\r\n]+" + token + "[ \t\r\n]*-->?",
                           PRegularExpression::Extended|PRegularExpression::IgnoreCase);
  PINDEX pos, len;
  while (text.FindRegEx(RegEx, pos, len))
    text.Splice(value, pos, len);
}


PString MyEndPoint::OnLoadEndPointStatus(const PString & htmlBlock)
{
  PINDEX i;
  PString substitution;

  for (PSafePtr<H323RegisteredEndPoint> ep = gatekeeperServer.GetFirstEndPoint(PSafeReadOnly); ep != NULL; ep++) {
    // make a copy of the repeating html chunk
    PString insert = htmlBlock;

    SpliceMacro(insert, "EndPointIdentifier", ep->GetIdentifier());

    PStringStream addresses;
    for (i = 0; i < ep->GetSignalAddressCount(); i++) {
      if (i > 0)
        addresses << "<br>";
      addresses << ep->GetSignalAddress(i);
    }
    SpliceMacro(insert, "CallSignalAddresses", addresses);

    PStringStream aliases;
    for (i = 0; i < ep->GetAliasCount(); i++) {
      if (i > 0)
        aliases << "<br>";
      aliases << ep->GetAlias(i);
    }
    SpliceMacro(insert, "EndPointAliases", aliases);

    PString str = "<i>Name:</i> " + ep->GetApplicationInfo();
    str.Replace("\t", "<BR><i>Version:</i> ");
    str.Replace("\t", " <i>Vendor:</i> ");
    SpliceMacro(insert, "Application", str);

    SpliceMacro(insert, "ActiveCalls", ep->GetCallCount());

    // Then put it into the page, moving insertion point along after it.
    substitution += insert;
  }

  return substitution;
}


PString MyEndPoint::OnLoadCallStatus(const PString & htmlBlock)
{
  PString substitution;

  for (PSafePtr<H323GatekeeperCall> call = gatekeeperServer.GetFirstCall(PSafeReadOnly); call != NULL; call++) {
    // make a copy of the repeating html chunk
    PString insert = htmlBlock;

    SpliceMacro(insert, "CallIdentifier",     call->GetCallIdentifier().AsString());
    SpliceMacro(insert, "EndPointIdentifier", call->GetEndPoint().GetIdentifier());
    SpliceMacro(insert, "SourceAddress",      call->GetSourceAddress());
    SpliceMacro(insert, "DestinationAddress", call->GetDestinationAddress());
    SpliceMacro(insert, "LastIRR",            call->GetLastInfoResponseTime().AsString("hh:mm:ss"));
    if (call->GetConnectedTime().GetTimeInSeconds() != 0)
      SpliceMacro(insert, "ConnectedTime",    call->GetConnectedTime().AsString("hh:mm:ss"));

    // Then put it into the page, moving insertion point along after it.
    substitution += insert;
  }

  return substitution;
}


PCREATE_SERVICE_MACRO_BLOCK(EndPointStatus,P_EMPTY,P_EMPTY,block)
{
  return OpenGk::Current().GetEndPoint().OnLoadEndPointStatus(block);
}


PCREATE_SERVICE_MACRO_BLOCK(CallStatus,P_EMPTY,P_EMPTY,block)
{
  return OpenGk::Current().GetEndPoint().OnLoadCallStatus(block);
}


///////////////////////////////////////////////////////////////

MyEndPoint::MyEndPoint(
#ifdef USE_OPAL
    OpalManager & manager
#endif
                       )
  :
#ifdef USE_OPAL
    H323EndPoint(manager),
#endif
#ifdef _MSC_VER
#pragma warning(disable:4355)
#endif
    gatekeeperServer(*this)
#ifdef _MSC_VER
#pragma warning(default:4355)
#endif
{
  terminalType = e_GatekeeperWithAudioMP;
}


MyEndPoint::~MyEndPoint()
{
  ClearAllCalls();
}


bool MyEndPoint::Initialise(PConfig & cfg, PConfigPage * rsrc)
{
  PINDEX i;

  // Set up software codecs.
  if (capabilities.GetSize() == 0) {
    AddAllCapabilities(0, 0, OpalGSM0610);
    AddAllCapabilities(0, 0, OpalG711uLaw);
    AddAllCapabilities(0, 0, OpalG711ALaw);

    H323_UserInputCapability::AddAllCapabilities(capabilities, 0, 1);
    PSYSTEMLOG(Info, "Main\tCodecs (in preference order):\n" << setprecision(2) << capabilities);
  }

  // Local alias name for H.323 endpoint
  SetLocalUserName(cfg.GetString(LocalUserNameKey, PProcess::Current().GetName()));
  rsrc->Add(new PHTTPStringField(LocalUserNameKey, 25, GetLocalUserName()));

  // IP TOS byte
  unsigned tos = cfg.GetInteger(TypeOfServiceKey, GetRtpIpTypeofService());
  SetRtpIpTypeofService(tos);
  rsrc->Add(new PHTTPIntegerField(TypeOfServiceKey, 0, 255, tos));

  // Interfaces to listen on
  PStringArray interfaces;
  PINDEX arraySize = cfg.GetInteger(GATEWAY_LISTENERS_SECTION, LISTENER_INTERFACE_KEY" Array Size");
  if (arraySize == 0)
    StartListener("*:*");
  else {
    for (i = 0; i < arraySize; i++)
      interfaces.Append(new H323TransportAddress(cfg.GetString(GATEWAY_LISTENERS_SECTION, psprintf(LISTENER_INTERFACE_KEY" %u", i+1), "")));
    StartListeners(interfaces);
  }
  rsrc->Add(new PHTTPFieldArray(new PHTTPStringField(GATEWAY_LISTENERS_SECTION"\\"LISTENER_INTERFACE_KEY, "Gateway Interfaces", 20), FALSE));

  if (listeners.IsEmpty()) {
    PSYSTEMLOG(Fatal, "Main\tCould not open H.323 Listener");
    return FALSE;
  }

  return gatekeeperServer.Initialise(cfg, rsrc);
}


H323Connection * MyEndPoint::CreateConnection(unsigned /*callReference*/)
{
  // Need to add a creation of a connection that can do gatekeeper routing.
  return NULL;
}


///////////////////////////////////////////////////////////////

MyGatekeeperServer::MyGatekeeperServer(MyEndPoint & ep)
  : H323GatekeeperServer(ep),
    endpoint(ep)
{
#ifdef H323_TRANSNEXUS_OSP
  ospProvider = NULL;
#endif
}


bool MyGatekeeperServer::Initialise(PConfig & cfg, PConfigPage * rsrc)
{
  PINDEX i;

  PWaitAndSignal mutex(reconfigurationMutex);

#ifdef H323_H501

  // set clearing house address
  PString clearingHouse = cfg.GetString(ClearingHouseKey);
  rsrc->Add(new PHTTPStringField(ClearingHouseKey, 25, clearingHouse));

  PString h501Interface = cfg.GetString(H501InterfaceKey);
  rsrc->Add(new PHTTPStringField(H501InterfaceKey, 25, h501Interface));

  if (clearingHouse.IsEmpty())
    SetPeerElement(NULL);
  else {
    if (!h501Interface)
      CreatePeerElement(h501Interface);
    if (!OpenPeerElement(clearingHouse)) {
      PSYSTEMLOG(Error, "Main\tCould not open clearing house at: " << clearingHouse);
    }
  }
#endif

#ifdef H323_TRANSNEXUS_OSP
  PString oldOSPServer = ospRoutingURL;
  ospRoutingURL = cfg.GetString(OSPRoutingURLKey, ospRoutingURL);
  rsrc->Add(new PHTTPStringField(OSPRoutingURLKey, 25, ospRoutingURL));
  bool ospChanged = oldOSPServer != ospRoutingURL;

  ospPrivateKeyFileName = cfg.GetString(OSPPrivateKeyFileKey, ospPrivateKeyFileName);
  rsrc->Add(new PHTTPStringField(OSPPrivateKeyFileKey, 25, ospPrivateKeyFileName));

  ospPublicKeyFileName = cfg.GetString(OSPPublicKeyFileKey, ospPublicKeyFileName);
  rsrc->Add(new PHTTPStringField(OSPPublicKeyFileKey, 25, ospPublicKeyFileName));

  ospServerKeyFileName = cfg.GetString(OSPServerKeyFileKey, ospServerKeyFileName);
  rsrc->Add(new PHTTPStringField(OSPServerKeyFileKey, 25, ospServerKeyFileName));

  if (!ospRoutingURL.IsEmpty()) {
    if (ospProvider != NULL && ospProvider->IsOpen() && ospChanged) {
      delete ospProvider;
      ospProvider = NULL;
    }
    ospProvider = new OpalOSP::Provider();
    int status;
    if (ospPrivateKeyFileName.IsEmpty() && ospPublicKeyFileName.IsEmpty() && ospServerKeyFileName.IsEmpty())
      status = ospProvider->Open(ospRoutingURL);
    else
      status = ospProvider->Open(ospRoutingURL, ospPrivateKeyFileName, ospPublicKeyFileName, ospServerKeyFileName);
    if (status != 0) {
      delete ospProvider;
      ospProvider = NULL;
    }
  } 
  
  else if (ospProvider != NULL) {
    ospProvider->Close();
    delete ospProvider;
    ospProvider = NULL;
  }

#endif

  PString gkid = cfg.GetString(GatekeeperIdentifierKey, OpenGk::Current().GetName() + " on " + PIPSocket::GetHostName());
  rsrc->Add(new PHTTPStringField(GatekeeperIdentifierKey, 25, gkid));
  SetGatekeeperIdentifier(gkid);

  // Interfaces to listen on
  H323TransportAddressArray interfaces;
  PINDEX arraySize = cfg.GetInteger(GATEKEEPER_LISTENERS_SECTION, LISTENER_INTERFACE_KEY" Array Size");
  for (i = 0; i < arraySize; i++)
    interfaces.Append(new H323TransportAddress(cfg.GetString(GATEKEEPER_LISTENERS_SECTION, psprintf(LISTENER_INTERFACE_KEY" %u", i+1), "")));
  AddListeners(interfaces);
  rsrc->Add(new PHTTPFieldArray(new PHTTPStringField(GATEKEEPER_LISTENERS_SECTION"\\"LISTENER_INTERFACE_KEY, "Gatekeeper Interfaces", 20), FALSE));

  SetAvailableBandwidth(cfg.GetInteger(AvailableBandwidthKey, GetAvailableBandwidth()/10)*10);
  rsrc->Add(new PHTTPIntegerField(AvailableBandwidthKey, 1, INT_MAX, GetAvailableBandwidth()/10, "kb/s"));

  defaultBandwidth = cfg.GetInteger(DefaultBandwidthKey, defaultBandwidth/10)*10;
  rsrc->Add(new PHTTPIntegerField(DefaultBandwidthKey, 1, INT_MAX, defaultBandwidth/10, "kb/s"));

  maximumBandwidth = cfg.GetInteger(MaximumBandwidthKey, maximumBandwidth/10)*10;
  rsrc->Add(new PHTTPIntegerField(MaximumBandwidthKey, 1, INT_MAX, maximumBandwidth/10, "kb/s"));

  defaultTimeToLive = cfg.GetInteger(DefaultTimeToLiveKey, defaultTimeToLive);
  rsrc->Add(new PHTTPIntegerField(DefaultTimeToLiveKey, 10, 86400, defaultTimeToLive, "seconds"));

  defaultInfoResponseRate = cfg.GetInteger(CallHeartbeatTimeKey, defaultInfoResponseRate);
  rsrc->Add(new PHTTPIntegerField(CallHeartbeatTimeKey, 0, 86400, defaultInfoResponseRate, "seconds"));

  overwriteOnSameSignalAddress = cfg.GetBoolean(OverwriteOnSameSignalAddressKey, overwriteOnSameSignalAddress);
  rsrc->Add(new PHTTPBooleanField(OverwriteOnSameSignalAddressKey, overwriteOnSameSignalAddress));

  canHaveDuplicateAlias = cfg.GetBoolean(CanHaveDuplicateAliasKey, canHaveDuplicateAlias);
  rsrc->Add(new PHTTPBooleanField(CanHaveDuplicateAliasKey, canHaveDuplicateAlias));

  canOnlyCallRegisteredEP = cfg.GetBoolean(CanOnlyCallRegisteredEPKey, canOnlyCallRegisteredEP);
  rsrc->Add(new PHTTPBooleanField(CanOnlyCallRegisteredEPKey, canOnlyCallRegisteredEP));

  canOnlyAnswerRegisteredEP = cfg.GetBoolean(CanOnlyAnswerRegisteredEPKey, canOnlyAnswerRegisteredEP);
  rsrc->Add(new PHTTPBooleanField(CanOnlyAnswerRegisteredEPKey, canOnlyAnswerRegisteredEP));

  answerCallPreGrantedARQ = cfg.GetBoolean(AnswerCallPreGrantedARQKey, answerCallPreGrantedARQ);
  rsrc->Add(new PHTTPBooleanField(AnswerCallPreGrantedARQKey, answerCallPreGrantedARQ));

  makeCallPreGrantedARQ = cfg.GetBoolean(MakeCallPreGrantedARQKey, makeCallPreGrantedARQ);
  rsrc->Add(new PHTTPBooleanField(MakeCallPreGrantedARQKey, makeCallPreGrantedARQ));

  aliasCanBeHostName = cfg.GetBoolean(AliasCanBeHostNameKey, aliasCanBeHostName);
  rsrc->Add(new PHTTPBooleanField(AliasCanBeHostNameKey, aliasCanBeHostName));

  isGatekeeperRouted = cfg.GetBoolean(IsGatekeeperRoutedKey, isGatekeeperRouted);
  rsrc->Add(new PHTTPBooleanField(IsGatekeeperRoutedKey, isGatekeeperRouted));

  routes.RemoveAll();
  arraySize = cfg.GetInteger(ROUTES_SECTION, ROUTES_KEY" Array Size");
  for (i = 0; i < arraySize; i++) {
    cfg.SetDefaultSection(psprintf(ROUTES_SECTION"\\"ROUTES_KEY" %u", i+1));
    RouteMap map(cfg.GetString(RouteAliasKey), cfg.GetString(RouteHostKey));
    if (map.IsValid())
      routes.Append(new RouteMap(map));
  }

  PHTTPCompositeField * routeFields = new PHTTPCompositeField(ROUTES_SECTION"\\"ROUTES_KEY" %u\\", "Alias Route Maps");
  routeFields->Append(new PHTTPStringField(RouteAliasKey, 20));
  routeFields->Append(new PHTTPStringField(RouteHostKey, 30));
  rsrc->Add(new PHTTPFieldArray(routeFields, TRUE));

  requireH235 = cfg.GetBoolean(RequireH235Key, requireH235);
  rsrc->Add(new PHTTPBooleanField(RequireH235Key, requireH235));

  passwords.RemoveAll();
  arraySize = cfg.GetInteger(USERS_SECTION, USERS_KEY" Array Size");
  for (i = 0; i < arraySize; i++) {
    cfg.SetDefaultSection(psprintf(USERS_SECTION"\\"USERS_KEY" %u", i+1));
    PString username = cfg.GetString(UsernameKey);
    if (!username) {
      PString password = PHTTPPasswordField::Decrypt(cfg.GetString(PasswordKey));
      passwords.SetAt(username, password);
    }
  }

  PHTTPCompositeField * security = new PHTTPCompositeField(USERS_SECTION"\\"USERS_KEY" %u\\", "Authentication Credentials");
  security->Append(new PHTTPStringField(UsernameKey, 25));
  security->Append(new PHTTPPasswordField(PasswordKey, 25));
  rsrc->Add(new PHTTPFieldArray(security, FALSE));

  return TRUE;
}


H323GatekeeperCall * MyGatekeeperServer::CreateCall(const OpalGloballyUniqueID & id,
                                                    H323GatekeeperCall::Direction dir)
{
  return new MyGatekeeperCall(*this, id, dir);
}


bool MyGatekeeperServer::TranslateAliasAddress(const H225_AliasAddress & alias,
                                               H225_ArrayOf_AliasAddress & aliases,
                                               H323TransportAddress & address,
                                               bool & isGkRouted,
                                               H323GatekeeperCall * call)
{
#ifdef H323_TRANSNEXUS_OSP
  if (ospProvider != NULL) {
    address = "127.0.0.1";
    return TRUE;
  }
#endif

  if (H323GatekeeperServer::TranslateAliasAddress(alias, aliases, address, isGkRouted, call))
    return TRUE;

  PString aliasString = H323GetAliasAddressString(alias);
  PTRACE(4, "Translating \"" << aliasString << "\" through route maps");

  PWaitAndSignal mutex(reconfigurationMutex);

  for (PINDEX i = 0; i < routes.GetSize(); i++) {
    PTRACE(4, "Checking route map " << routes[i]);
    if (routes[i].IsMatch(aliasString)) {
      address = routes[i].GetHost();
      PTRACE(3, "Translated \"" << aliasString << "\" to " << address);
      break;
    }
  }

  return TRUE;
}


MyGatekeeperServer::RouteMap::RouteMap(const PString & theAlias, const PString & theHost)
  : alias(theAlias),
    regex('^' + theAlias + '$'),
    host(theHost)
{
}


void MyGatekeeperServer::RouteMap::PrintOn(ostream & strm) const
{
  strm << '"' << alias << "\" => " << host;
}


bool MyGatekeeperServer::RouteMap::IsValid() const
{
  return !alias && regex.GetErrorCode() == PRegularExpression::NoError && !host;
}


bool MyGatekeeperServer::RouteMap::IsMatch(const PString & alias) const
{
  PINDEX start;
  return regex.Execute(alias, start);
}


///////////////////////////////////////////////////////////////

MyGatekeeperCall::MyGatekeeperCall(MyGatekeeperServer & gk,
                                   const OpalGloballyUniqueID & id,
                                   Direction dir)
  : H323GatekeeperCall(gk, id, dir)
{
#ifdef H323_TRANSNEXUS_OSP
  ospTransaction = NULL;
#endif
}

#ifdef H323_TRANSNEXUS_OSP
static bool GetE164Alias(const H225_ArrayOf_AliasAddress & aliases, H225_AliasAddress & destAlias)
{
  PINDEX i;
  for (i = 0; i < aliases.GetSize(); ++i) {
    if (aliases[i].GetTag() == H225_AliasAddress::e_dialedDigits) {
      destAlias = aliases[i];
      return TRUE;
    }
  }
  return FALSE;
}

bool MyGatekeeperCall::AuthoriseOSPCall(H323GatekeeperARQ & info)
{
  int result;

  // if making call, authorise the call and insert the token
  if (!info.arq.m_answerCall) {

    OpalOSP::Transaction::AuthorisationInfo authInfo;

    // get the source call signalling address
    if (info.arq.HasOptionalField(H225_AdmissionRequest::e_srcCallSignalAddress))
      authInfo.ospvSource = OpalOSP::TransportAddressToOSPString(info.arq.m_srcCallSignalAddress);
    else
      authInfo.ospvSource = OpalOSP::TransportAddressToOSPString(info.endpoint->GetSignalAddress(0));

    // get the source number
    if (!GetE164Alias(info.arq.m_srcInfo, authInfo.callingNumber)) {
      PTRACE(1, "OSP\tNo E164 source address in ARQ");
      info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_incompleteAddress);
      return FALSE;
    }

    // get the dest number
    if (!info.arq.HasOptionalField(H225_AdmissionRequest::e_destinationInfo)) {
      PTRACE(1, "OSP\tNo dest aliases in ARQ");
      info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_incompleteAddress);
      return FALSE;
    }
    if (!GetE164Alias(info.arq.m_destinationInfo, authInfo.calledNumber)) {
      PTRACE(1, "OSP\tNo E164 source address in ARQ");
      info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_incompleteAddress);
      return FALSE;
    }

    // get the call ID
    authInfo.callID           = this->GetCallIdentifier();

    // authorise the call
    unsigned numberOfDestinations = 1;
    if ((result = ospTransaction->Authorise(authInfo, numberOfDestinations)) != 0) {
      PTRACE(1, "OSP\tCannot authorise ARQ - result = " << result);
      info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_incompleteAddress);
      return FALSE;
    } 
    if (numberOfDestinations == 0) {
      PTRACE(1, "OSP\tNo destinations available");
      info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_noRouteToDestination);
      return FALSE;
    }

    // get the destination
    OpalOSP::Transaction::DestinationInfo destInfo;
    if ((result = ospTransaction->GetFirstDestination(destInfo)) != 0) {
      PTRACE(1, "OSP\tCannot get first destination - result = " << result);
      info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_undefinedReason);
      return FALSE;
    } 

    // insert destination into the ACF
    if (!destInfo.Insert(info.acf)) {
      PTRACE(1, "OSP\tCannot insert information info ACF");
      info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_undefinedReason);
      return FALSE;
    }

    PTRACE(4, "OSP routed call to " << destInfo.calledNumber << "@" << destInfo.destinationAddress);
    return TRUE;
  }

  // if answering call, validate the token
  OpalOSP::Transaction::ValidationInfo valInfo;

  // get the token
  if (!info.arq.HasOptionalField(H225_AdmissionRequest::e_tokens) || 
     !valInfo.ExtractToken(info.arq.m_tokens)) {
    PTRACE(1, "OSP\tNo tokens in in ARQ");
    info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_invalidPermission);
    return FALSE;
  }

  // get the source call signalling address
  if (info.arq.HasOptionalField(H225_AdmissionRequest::e_srcCallSignalAddress))
    valInfo.ospvSource = OpalOSP::TransportAddressToOSPString(info.arq.m_srcCallSignalAddress);
  else
    valInfo.ospvSource = OpalOSP::TransportAddressToOSPString(info.endpoint->GetSignalAddress(0));

  // get the dest call signalling address
  if (info.arq.HasOptionalField(H225_AdmissionRequest::e_destCallSignalAddress))
    valInfo.ospvDest = OpalOSP::TransportAddressToOSPString(info.arq.m_destCallSignalAddress);
  else
    valInfo.ospvDest = OpalOSP::TransportAddressToOSPString(info.endpoint->GetSignalAddress(0));

  // get the source number
  if (!GetE164Alias(info.arq.m_srcInfo, valInfo.callingNumber)) {
    PTRACE(1, "OSP\tNo E164 source address in ARQ");
    info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_incompleteAddress);
    return FALSE;
  }

  // get the dest number
  if (!info.arq.HasOptionalField(H225_AdmissionRequest::e_destinationInfo)) {
    PTRACE(1, "OSP\tNo dest aliases in ARQ");
    info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_incompleteAddress);
    return FALSE;
  }
  if (!GetE164Alias(info.arq.m_destinationInfo, valInfo.calledNumber)) {
    PTRACE(1, "OSP\tNo E164 source address in ARQ");
    info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_incompleteAddress);
    return FALSE;
  }

  // get the call ID
  valInfo.callID = this->GetCallIdentifier();

  // validate the token
  bool authorised;
  unsigned timeLimit;
  if ((result = ospTransaction->Validate(valInfo, authorised, timeLimit)) != 0) {
    PTRACE(1, "OSP\tCannot validate ARQ - result = " << result);
    info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_invalidPermission);
    return FALSE;
  }

  if (!authorised) {
    PTRACE(1, "OSP\tCall not authorised");
    info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_requestDenied);
    return FALSE;
  }

  PTRACE(4, "OSP authorised call with time limit of " << timeLimit);
  return TRUE;
}

#endif

H323GatekeeperRequest::Response MyGatekeeperCall::OnAdmission(H323GatekeeperARQ & info)
{
#ifdef TEST_TOKEN
  info.acf.IncludeOptionalField(H225_AdmissionConfirm::e_tokens);
  info.acf.m_tokens.SetSize(1);
  info.acf.m_tokens[0].m_tokenOID = "1.2.36.76840296.1";
  info.acf.m_tokens[0].IncludeOptionalField(H235_ClearToken::e_nonStandard);
  info.acf.m_tokens[0].m_nonStandard.m_nonStandardIdentifier = "1.2.36.76840296.1.1";
  info.acf.m_tokens[0].m_nonStandard.m_data = "SnfYt0jUuZ4lVQv8umRYaH2JltXDRW6IuYcnASVU";
#endif

#ifdef H323_TRANSNEXUS_OSP
  OpalOSP::Provider * ospProvider = ((MyGatekeeperServer &)gatekeeper).GetOSPProvider();
  if (ospProvider != NULL) {
    // need to add RIP as clearing house is involved
    if (info.IsFastResponseRequired())
      return H323GatekeeperRequest::InProgress(30000);

    ospTransaction = new OpalOSP::Transaction();
    int result;
    if ((result = ospTransaction->Open(*ospProvider)) != 0) {
      delete ospTransaction;
      ospTransaction = NULL;
      PTRACE(1, "OSP\tCannot open transaction - result = " << result);
      info.arj.m_rejectReason.SetTag(H225_AdmissionRejectReason::e_exceedsCallCapacity);
      return H323GatekeeperRequest::Reject;
    }

    if (!AuthoriseOSPCall(info)) {
      delete ospTransaction;
      ospTransaction = NULL;
      return H323GatekeeperRequest::Reject;
    }

    return H323GatekeeperRequest::Confirm;
#endif

  return H323GatekeeperCall::OnAdmission(info);
}


MyGatekeeperCall::~MyGatekeeperCall()
{
#ifdef H323_TRANSNEXUS_OSP
  if (ospTransaction != NULL) {
    ospTransaction->SetEndReason(callEndReason);
    PTimeInterval duration;
    if (connectedTime.GetTimeInSeconds() != 0 && callEndTime.GetTimeInSeconds() != 0)
      duration = callEndTime - connectedTime;
    ospTransaction->CallEnd(callEndTime.GetTimeInSeconds());
    delete ospTransaction;
  }
#endif
}


// End of File ///////////////////////////////////////////////////////////////
